/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "input.h"
#include "genmap.h"
#include "ped.h"
#include <stdlib.h>
#include <string.h>

void computeChiSq(pedigree_t *ped, double *chiSq, int *dof);

/* Computes the chi-square statistics for markers in the pedigree and writes
   them to a file.
 */
int main(int argc, char **argv) {
    char        *input, *output;
    char        *eol = "\n";
    FILE        *fp;
    pedigree_t  ped;
    double      *chiSq;
    int         *dof;
    int         i;
    
    if (argc == 3 && strcmp(argv[1], "-d") != 0) {
        input = argv[1];
        output = argv[2];
    } else if (argc == 4 && strcmp(argv[1], "-d") == 0) {
        input = argv[2];
        output = argv[3];
        eol = "\r\n";
    } else {
        fprintf(stderr, "%s [-d] <input file> <output file>\n", *argv);
        return EXIT_FAILURE;
    }
    
    if (!readFile(&ped, input))
        return EXIT_FAILURE;
    
    chiSq = (double *)malloc(ped.numMarks * sizeof(double));
    dof = (int *)malloc(ped.numMarks * sizeof(int));
    computeChiSq(&ped, chiSq, dof);
   
    fp = fopen(output, "w");
    if (!fp) {
        fprintf(stderr, "Couldn't open file %s for writing\n", output);
        return EXIT_FAILURE;
    }
    
    fprintf(fp, "Marker\tDOF\tChi-sq%s", eol);
    for (i = 0; i < ped.numMarks; ++i)
        fprintf(fp, "%s\t%d\t%.3f%s", ped.markers[i].name, dof[i], chiSq[i],
                eol);
    
    fclose(fp);
    free(chiSq);
    free(dof);
    freePed(&ped);
    
    return EXIT_SUCCESS;
}

/* Computes the Chi-square of each marker's segregation. This computation does
   not make sense for general IBD data, but only for mutually exclusive
   segregation classes. The function tries to find minimal genotypes and
   compute the Chi-square of these. If there are any markers which are not
   minimal and are not unknown, they are treated as unknowns and a message is
   printed to standard out.
 */
void computeChiSq(pedigree_t *ped, double *chiSq, int *dof)
{
    int     i, j;
    
    for (i = 0; i < ped->numMarks; ++i) {
        int     counts[16];
        int     valid[16]; /* part of segregation class */
        int     skipped = 0; /* number of ambiguous genotypes */
        int     mask = 0; /* which IBDs have been seen so far */
        int     numKnown = 0;
        double  expProp[15] = {0.0, 0.25, 0.25, 0.5, 0.25, 0.5, 0.5, 0.75, 0.25,
                               0.5, 0.5, 0.75, 0.5, 0.75, 0.75};
        
        dof[i] = -1;
        chiSq[i] = 0.0;
        
        for (j = 0; j < 16; ++j)
            counts[j] = valid[j] = 0;
        
        for (j = 0; j < ped->numInds; ++j)
            ++counts[ped->markers[i].obs[j]];
        
        for (j = 1; j < 15; ++j) {
            if (mask & j)
                skipped += counts[j];
            else if (counts[j]) {
                mask |= j;
                numKnown += counts[j];
                valid[j] = TRUE;
            }
        }
        
        if (skipped)
            fprintf(stderr,
                    "Warning: ignoring %d ambiguous genotypes in marker %s\n",
                    skipped, ped->markers[i].name);
        
        /* If one segregation class is completely missing, it should still be
           counted in the DOF and the Chi-square. */
        valid[OBSERVATION_UNKNOWN & ~mask] = TRUE;
        
        for (j = 1; j < 15; ++j) {
            if (valid[j]) {
                double  expected = expProp[j] * numKnown;
                double  diff = counts[j] - expected;
                
                ++dof[i];
                chiSq[i] += diff * diff / expected;
            }
        }
    }
}
