/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "improve.h"
#include "mapLlhd.h"
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

/* These parameters control the actual Monte Carlo algorithm */
#define NUM_STEPS_PER_MARKER 200
#define STEPS_PER_MARKER_WITHOUT_IMPROVE 50

/* These parameters describe probability distribution over the neighborhood */
#define PROB_FLIP 0.50
#define DECAY_FLIP_LEN 0.9
#define DECAY_MOVE_DIST 0.9

void permuteMarkers(map_t *, const map_t *);
int tryPermuteMarkers(map_t *newMap, const map_t *oldMap);

void initRand();
int randInt(int);
int randIntDecay(int, double decay);

#ifdef NO_RAND48_FAMILY
static double drand48()
{
    return rand() / RAND_MAX;
}

#define srand48 srand
#define lrand48 rand
#define LRAND48_MAX     RAND_MAX
#else
#define LRAND48_MAX     (1UL << 31)
#endif

/* Improves the marker order in the map using a stochastic Monte Carlo search.
 */
void improveMap(map_t *map)
{
    map_t   curMap, newMap;
    int     i, j;
    int     maxStepsWithoutImprove;
    double  temps[] = {0.5, 0.05};
    
    /* Maps with fewer than 4 markers cannot be improved because all the orders
       have the same log likelihood.
     */
    if (map->numMarks <= 3)
        return;
    
    initRand();
    
    allocSimilarMap(&curMap, map);
    allocSimilarMap(&newMap, map);
    
    fprintf(stderr, "initial log likelihood = %.2f\n", map->logLlhd);
    
    maxStepsWithoutImprove = STEPS_PER_MARKER_WITHOUT_IMPROVE * map->numMarks;
    for (j = 0; j < sizeof(temps) / sizeof(*temps); ++j) {
        double  temp = temps[j];
        int     lastImprove = 0;
        
        copyMap(&curMap, map);
        
        for (i = 0; i < NUM_STEPS_PER_MARKER * map->numMarks; ++i) {
            prob_t  diffLogLlhd;
            
            permuteMarkers(&newMap, &curMap);
            computeMapLlhd(&newMap);
            
            diffLogLlhd = newMap.logLlhd - map->logLlhd;
            if (diffLogLlhd > 0 || drand48() < exp(diffLogLlhd / temp)) {
                swapMaps(&curMap, &newMap);
                if (curMap.logLlhd > map->logLlhd) {
                    copyMap(map, &curMap);
                    fprintf(stderr, "i = %d, log likelihood = %.1f\n", i,
                            map->logLlhd);
                    lastImprove = i;
                }
            }
            
            if (i >= lastImprove + maxStepsWithoutImprove) {
                fprintf(stderr,"i = %d, reseting\n", i);
                copyMap(&curMap, map);
                lastImprove = i;
            }
        }
    }
    
    freeMap(&curMap);
    freeMap(&newMap);
}

/* Randomly permutes the markers in oldMap to create newMap. */
void permuteMarkers(map_t *newMap, const map_t *oldMap)
{
    while (!tryPermuteMarkers(newMap, oldMap))
        ;
}

/* Randomly permutes the markers in oldMap to create newMap or fails. This
   allows the function to deal with out-of-bound parameters
   non-deterministically, i.e. by failing, at which point another permutation
   will be tried. */
int tryPermuteMarkers(map_t *newMap, const map_t *oldMap)
{
    if (drand48() < PROB_FLIP) {
        /* flip a subsection of the map in place */
        int     len = randIntDecay(oldMap->numMarks - 3, DECAY_FLIP_LEN) + 2;
        int     ndx = randInt(oldMap->numMarks - len + 1);
        int     i;
        
        copyMap(newMap, oldMap);
        
        for (i = 0; i < len; ++i)
            newMap->order[ndx + i] = oldMap->order[ndx + len - i - 1];
    } else {
        /* move a marker from one position to another. */
        int     src = randInt(oldMap->numMarks);
        int     dist = randIntDecay(oldMap->numMarks - 1, DECAY_MOVE_DIST);
        int     dst;
        
        if (randInt(2)) {
            dst = src + dist;
            if (dst >= oldMap->numMarks)
                return FALSE;
        } else {
            dst = src - dist;
            if (dst < 0)
                return FALSE;
        }
        
        newMap->numMarks = oldMap->numMarks;
        if (src < dst) {
            memcpy(newMap->order, oldMap->order, src * sizeof(marker_t *));
            memcpy(newMap->order + src, oldMap->order + src + 1,
                   (dst - src) * sizeof(marker_t *));
            newMap->order[dst] = oldMap->order[src];
            memcpy(newMap->order + dst + 1, oldMap->order + dst + 1,
                   (oldMap->numMarks - dst - 1) * sizeof(marker_t *));
        } else {
            memcpy(newMap->order, oldMap->order, dst * sizeof(marker_t *));
            newMap->order[dst] = oldMap->order[src];
            memcpy(newMap->order + dst + 1, oldMap->order + dst,
                   (src - dst) * sizeof(marker_t *));
            memcpy(newMap->order + src + 1, oldMap->order + src + 1,
                   (oldMap->numMarks - src - 1) * sizeof(marker_t *));
        }
    }
    
    return TRUE;
}

void initRand()
{
    srand48(time(NULL));
}

int randInt(int limit)
{
    unsigned long   r;
    unsigned long   thrsh = LRAND48_MAX - LRAND48_MAX % (unsigned long) limit;
    
    while ((r = lrand48()) >= thrsh)
        ;
    return r / (LRAND48_MAX / limit);
}

/* Returns a random integer in the range [0, limit), but distributed so that
   the probability of n is proportional to decay ** n. The parameter decay must
   be in the range (0.0, 1.0].
 */
int randIntDecay(int limit, double decay)
{
    double  d;
    double  p = decay;
    int     i;
    
    if (decay == 1.0)
        return randInt(limit);
    
    for (i = 0; i < limit - 1; ++i)
        p *= decay;
    
    d = 1.0 - drand48() / (1.0 - p);
    
    for (i = 0; i < limit - 1; ++i) {
        if (d >= decay)
            break;
        d /= decay;
    }
    return i;
}
