/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "input.h"
#include "ped.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>

/* This constraint is inherent in the way that the symbols are entered into
   the table.
   
   This is because ReadSegType itself inserts 2 symbols, and
   ReadSegTypeLongCross inserts at most 17 symbols.  If the definitions of any
   of these are changed to change these characteristics, then this constant
   should also be changed.
 */
#define MAX_NUM_SYMBOLS 19

#define FILE_TYPE_UNKNOWN -1
#define FILE_TYPE_BACKCROSS 0
#define FILE_TYPE_INTERCROSS 1
#define FILE_TYPE_OUTBRED 2

#define BUF_SIZE 256
#define MAX_NAME_LEN 16
#define MAX_ALLELE_LEN 16

/* Represents a symbol used in the "outbred" file type. An arbitrary string of
   characters (name) is associated with a genotype (val).
 */
typedef struct {
    char        *name;
    obs_t       val;
    int         fixed;
} outbredSymbol_t;

typedef struct {
    char    *str;
    int     fileType;
} fileDesc_t;

int readF2File(pedigree_t *ped, obs_t *obs, FILE *fp, int fileType);

int readOutbredFile(pedigree_t *ped, obs_t *obs, FILE *fp);
int ReadOutbredMarker(FILE *fp, obs_t *obs, int numInds,
                      const char *markName);
int ReadSegType(FILE *fp, outbredSymbol_t *syms, int *pNumSyms,
                const char *markName);
int ReadSegTypeCross(FILE *fp, outbredSymbol_t *syms, int *pNumSyms,
                      const char *markName);
int ReadSegTypeLongCross(FILE *fp, outbredSymbol_t *syms, int *pNumSyms,
                         const char *markName);
void AddLongCrossGenotype(outbredSymbol_t *symbs, int *pCount, char *allele1,
                          char *allele2, char sep, int gen);
int ReadSegTypeExplicit(FILE *fp, outbredSymbol_t *syms, int *pNumSyms,
                         const char *markName);
int ReadSegTypeExplicitOld(FILE *fp, outbredSymbol_t *syms, int *pNumSyms,
                            const char *markName);
void AddSymbolGenotype(outbredSymbol_t *symbs, int *pCount, const char *name,
                       int gen);
outbredSymbol_t *LookupOutbredSymbol(outbredSymbol_t *symbs, int count,
                                     const char *name);
void FreeOutbredSymbols(outbredSymbol_t *, int numSymbs);
int ReadToken(char *buf, int bufSz, FILE *fp);
void ReadToEol(FILE *fp);

/* Tries to read pedigree information from the given file. Returns a flag
   indicating whether or not it succeeded.
 */
int readFile(pedigree_t *ped, const char *fname)
{
    FILE        *fp = fopen(fname, "r");
    char        buf[BUF_SIZE];
    char        *p, *endp;
    int         numMrk, numInd;
    obs_t       *obs;
    const char  *headerLine = "data type ";
    int         fileType = FILE_TYPE_UNKNOWN;
    int         i;
    fileDesc_t  descs[] = {{"f2 backcross", FILE_TYPE_BACKCROSS},
                           {"F2 backcross", FILE_TYPE_BACKCROSS},
                           {"f2 intercross", FILE_TYPE_INTERCROSS},
                           {"F2 intercross", FILE_TYPE_INTERCROSS},
                           {"outbred", FILE_TYPE_OUTBRED}};
    int         numDescs = sizeof(descs) / sizeof(fileDesc_t);
    
    if (fp == NULL) {
        fprintf(stderr, "Couldn't open file %s for reading\n", fname);
        return FALSE;
    }
    
    fgets(buf, strlen(headerLine) + 1, fp);
    if (strcmp(buf, headerLine) != 0) {
        fprintf(stderr, "Unknown type of data file\n");
        fprintf(stderr, "First line should begin with: %s\n", headerLine);
        fclose(fp);
        return FALSE;
    }
    
    fgets(buf, BUF_SIZE, fp);
    for (i = 0; i < numDescs; ++i) {
        if (strncmp(buf, descs[i].str, strlen(descs[i].str)) == 0) {
            fileType = descs[i].fileType;
            break;
        }
    }
    if (fileType == FILE_TYPE_UNKNOWN) {
        fprintf(stderr, "Unknown file type in %s: %s\n", fname, buf);
        fclose(fp);
        return FALSE;
    }
    if (buf[strlen(buf) - 1] != '\n')
        ReadToEol(fp);
    
    fgets(buf, BUF_SIZE, fp);
    numInd = strtol(buf, &p, 10);
    numMrk = strtol(p, &endp, 10);
    if (p == buf || endp == p) {
        fprintf(stderr, "Error parsing second line of file %s: %s\n", fname,
                buf);
        fclose(fp);
        return FALSE;
    }
    if (buf[strlen(buf) - 1] != '\n')
        ReadToEol(fp);
    
    obs = (obs_t *)malloc(numMrk * numInd * sizeof(obs_t));
    ped->markers = (marker_t *)malloc(numMrk * sizeof(marker_t));
    ped->numMarks = numMrk;
    ped->numInds = numInd;
    
    if (fileType != FILE_TYPE_OUTBRED)
        return readF2File(ped, obs, fp, fileType);
    else
        return readOutbredFile(ped, obs, fp);
}


/* Reads a file in which each genotype is specified by a single character, as
   opposed to a token. Currently supports the "f2 intercross" and "f2 backcross"
   formats.
 */
int readF2File(pedigree_t *ped, obs_t *obs, FILE *fp, int fileType)
{
    obs_t   *lookupBase, *lookup;
    int     i;
    int     success = TRUE;
    
    lookupBase = (obs_t *) malloc((CHAR_MAX - CHAR_MIN + 1) * sizeof(obs_t));
    lookup = lookupBase - CHAR_MIN;
    
    for (i = CHAR_MIN; i <= CHAR_MAX; ++i)
        lookup[i] = 0x0;
    lookup['-'] = OBSERVATION_UNKNOWN;
    
    if (fileType == FILE_TYPE_BACKCROSS) {
        lookup['H'] = OBSERVATION_1010;
        lookup['A'] = OBSERVATION_0101;
    } else if (fileType == FILE_TYPE_INTERCROSS) {
        for (i = 1; i < 10; ++i)
            lookup['0' + i] = i;
        for (i = 0; i < 6; ++i)
            lookup['a' + i] = 10 + i;
        lookup['A'] = OBSERVATION_0001;
        lookup['H'] = OBSERVATION_0110;
        lookup['B'] = OBSERVATION_1000;
        lookup['C'] = OBSERVATION_1110;
        lookup['D'] = OBSERVATION_0111;
    }
    
    for (i = 0; i < ped->numMarks; ++i) {
        int         c;
        marker_t    *mark = ped->markers + i;
        int         j;
        
        mark->obs = obs + ped->numInds * i;
        
        if ((mark->name = readName(fp)) == NULL) {
            fprintf(stderr, "Warning: File supposed to have ");
            fprintf(stderr, "%d markers, but only has %d\n", ped->numMarks, i);
            ped->numMarks = i;
            break;
        }
        
        for (j = 0; j < ped->numInds; ++j) {
            while(isspace(c = getc(fp)) && c != '\n')
                ;
            
            if (c == EOF) {
                fprintf(stderr, "Error: Unexpected end of file\n");
                fclose(fp);
                free(lookupBase);
                freePed(ped);
                return FALSE;
            }
            
            if ((mark->obs[j] = lookup[c]) == 0x0) {
                success = FALSE;
                
                if (c == '\n') {
                    fprintf(stderr,
                            "Error: marker %s has only %d observations\n",
                            mark->name, j);
                    ungetc(c, fp);
                    break;
                }
                
                fprintf(stderr,
                        "Error: bad observation character in marker %s: '%c'\n",
                        mark->name, (char) c);
            }
        }
        
        while (isspace(c = getc(fp)) && c != '\n')
            ;
        if (c != '\n' && c != EOF) {
            fprintf(stderr, "Error: marker %s: expected end of line, got %c\n",
                    mark->name, (char) c);
            fclose(fp);
            free(lookupBase);
            ped->numMarks = i + 1;
            freePed(ped);
            return FALSE;
        }
    }
    
    free(lookupBase);
    fclose(fp);
    if (!success)
        freePed(ped);
    return success;
}

/* Reads the file in outbred format. Returns whether or not it succeeded. */
int readOutbredFile(pedigree_t *ped, obs_t *obs, FILE *fp)
{
    int     success = TRUE;
    int     i;
    
    for (i = 0; i < ped->numMarks; ++i) {
        marker_t    *mark = ped->markers + i;
        
        mark->name = readName(fp);
        if (mark->name == NULL) {
            fprintf(stderr, "Warning: File supposed to have ");
            fprintf(stderr, "%d markers, but only has %d\n", ped->numMarks, i);
            ped->numMarks = i;
            break;
        }
        mark->obs = obs + i * ped->numInds;
        
        if (!ReadOutbredMarker(fp, mark->obs, ped->numInds, mark->name))
            success = FALSE;
    }
    
    fclose(fp);
    if (!success)
        freePed(ped);
    return success;
}

/* Reads a marker name from the file stream. A marker name is a string of
   characters terminated by white space. The marker name may begin with an *,
   which is ignored, for compatibility with MapMaker/CarthaGene format files.
   
   After the function the stream position is immediately after the first white
   space character. If the there is no marker name before the end of the file,
   the function returns NULL.
   
   The caller is resposible for calling free on a non-NULL returned pointer.
 */
char *readName(FILE *fp)
{
    char    *name = (char *)malloc(MAX_NAME_LEN);
    int     c;
    int     i = 0;
    
    c = getc(fp);
    if (c == EOF) {
        free(name);
        return NULL;
    } else if (c != '*')
        name[i++] = c;
    
    while (!isspace(c = getc(fp)) && c != EOF) {
        if (i < MAX_NAME_LEN - 1)
            name[i] = c;
        else if (i == MAX_NAME_LEN - 1) {
            name[i] = '\0';
            fprintf(stderr, "Name beginning with %s too long, truncating\n",
                    name);
        }
        ++i;
    }
    
    ungetc(c, fp);
    
    if (i < MAX_NAME_LEN)
        name[i] = '\0';
    
    return name;
}

/* Reads one line describing a marker in outbred format which consists of:
   [*]<Marker name> <segregation type> <genotypes> ...
   where the genotypes are separated by white space (but not newlines). The
   marker name has already been read by the time this method is called.  The
   segregation type can take various formats. See the functions below.  Each
   genotype is a symbol whose meaning is encoded in the segregation type
   section, or is a symbol with a fixed meaning. These mappings are stored in
   the symbols array.
   
   The markName variable is only used for printing error messages.
   
   It returns TRUE iff it successfully read the marker line. Even if it fails,
   it reads to the end of the line so that the next marker can be read. */
int ReadOutbredMarker(FILE *fp, obs_t *obs, int numInds,
                      const char *markName)
{
    char            tok[BUF_SIZE];
    outbredSymbol_t symbols[MAX_NUM_SYMBOLS];
    int             numSymbs;
    int             j;
    obs_t           totalObs = 0x0; /* overall mask of all genotypes observed */
    int             success = TRUE;
    
    if (!ReadSegType(fp, symbols, &numSymbs, markName)) {
        FreeOutbredSymbols(symbols, numSymbs);
        ReadToEol(fp);
        return FALSE;
    }
    
    for (j = 0; j < numInds; ++j) {
        outbredSymbol_t *symb;
        
        if (!ReadToken(tok, BUF_SIZE, fp)) {
            fprintf(stderr, "Error: marker %s has only %d genotypes\n",
                    markName, j);
            success = FALSE;
            break;
        }
        
        symb = LookupOutbredSymbol(symbols, numSymbs, tok);
        if (symb == NULL) {
            fprintf(stderr, "Error: Unknown genotype ");
            fprintf(stderr, "%s used with marker %s, individual %d\n",
                    tok, markName, j + 1);
            success = FALSE;
        } else {
            obs[j] = symb->val;
            if (symb->val != OBSERVATION_UNKNOWN)
                totalObs = totalObs | symb->val;
        }
    }
    
    /* The following is simply a check for possible data entry errors. It prints
       a warning if not all possible genotypes have been entered. */
    if (success) {
        if (totalObs == 0)
            fprintf(stderr, "Warning: marker %s has all genotypes unknown\n",
                    markName);
        else if (totalObs != OBSERVATION_UNKNOWN) {
            obs_t   missingObs = ~totalObs & OBSERVATION_UNKNOWN;
            int     foundSymb = FALSE;
            
            fprintf(stderr, "Warning: marker %s missing ", markName);
            
            for (j = 0; j < numSymbs; ++j) {
                if (symbols[j].val == missingObs) {
                    if (!foundSymb) {
                        fprintf(stderr, "genotype:");
                        foundSymb = TRUE;
                    }
                    fprintf(stderr, " %s", symbols[j].name);
                }
            }
            
            if (!foundSymb) {
                fprintf(stderr, "multiple genotypes:");
                for (j = 0; j < numSymbs; ++j) {
                    if ((symbols[j].val & totalObs) == 0)
                        fprintf(stderr, " %s", symbols[j].name);
                }
            }
            fprintf(stderr, "\n");
        }
    }
    
    FreeOutbredSymbols(symbols, numSymbs);
    ReadToEol(fp);
    return success;
}

/* Reads the segregation type from the stream. It uses the segregation type
   to fill in the symbols structures which provide token/phenotype->genotype
   mappings. The markName paramter is used only in order to provide more
   informative error messages. Returns false if it could not parse the
   segregation type. Whether or not it succeeds, the caller is responsible for
   calling FreeOutbredSymbols on the symbols.
   
   This funtion adds two symbols to the table, plus however may are added
   by the subroutines.
 */
int ReadSegType(FILE *fp, outbredSymbol_t *symbols, int *pNumSymbs,
                 const char *markName)
{
    int         c;
    
    symbols[0].name = "-";
    symbols[0].val = OBSERVATION_UNKNOWN;
    symbols[0].fixed = TRUE;
    symbols[1].name = "--";
    symbols[1].val = OBSERVATION_UNKNOWN;
    symbols[1].fixed = TRUE;
    *pNumSymbs = 2;
    
    while (isspace(c = getc(fp)))
        ;
    
    switch(c) {
        case '<':
            if ((c = getc(fp)) == '<')
                return ReadSegTypeLongCross(fp, symbols, pNumSymbs, markName);
            else {
                ungetc(c, fp);
                return ReadSegTypeCross(fp, symbols, pNumSymbs, markName);
            }
        case '[':
            return ReadSegTypeExplicit(fp, symbols, pNumSymbs, markName);
        default:
            ungetc(c, fp);
            return ReadSegTypeExplicitOld(fp, symbols, pNumSymbs, markName);
    }
        
}

/* Reads the segregation type in the format: <XXxXX> where the X's denote
   arbitrary characters. At the point when this function has been called, the
   first character ('<') has already been parsed.
   
   This function adds a maximum of 16 symbols to the table.
 */
int ReadSegTypeCross(FILE *fp, outbredSymbol_t *symbols, int *pNumSymbs,
                      const char *markName)
{
    char    buf[6];
    int     i;
    int     nullParentAllele;
    
    /* The meaning of the positions in buf are:
        0 - mother haplotype
        1 - mother haplotype
        2 - 'x'
        3 - father haplotype
        4 - father haplotype
        5 - '>'
     */
    fread(buf, 1, 6, fp);
    if (buf[2] != 'x' || buf[5] != '>') {
        fprintf(stderr, "Error: segregation type for marker %s not", markName);
        fprintf(stderr, " properly formatted: <");
        fwrite(buf, 1, 6, stderr);
        putc('\n', stderr);
        return FALSE;
    }
    
    nullParentAllele = buf[0] == '0' || buf[1] == '0'
                         || buf[3] == '0' || buf[4] == '0';
    
    for (i = 0; i < 4; ++i) {
        char    cMoth = buf[i & 0x01];
        char    cFath = buf[3 + (i >> 1)];
        char    symb[3];
        
        symb[0] = cMoth;
        symb[1] = cFath;
        symb[2] = '\0';
        AddSymbolGenotype(symbols, pNumSymbs, symb, i);
        
        if (!nullParentAllele || cFath == cMoth || cFath == '0') {
            symb[1] = '-';
            AddSymbolGenotype(symbols, pNumSymbs, symb, i);
        }
        symb[1] = '?';
        AddSymbolGenotype(symbols, pNumSymbs, symb, i);
        
        symb[0] = cFath;
        symb[1] = cMoth;
        AddSymbolGenotype(symbols, pNumSymbs, symb, i);
        
        if (!nullParentAllele || cFath == cMoth || cMoth == '0') {
            symb[1] = '-';
            AddSymbolGenotype(symbols, pNumSymbs, symb, i);
        }
        symb[1] = '?';
        AddSymbolGenotype(symbols, pNumSymbs, symb, i);
    }
    
    return TRUE;
}

/* Reads a segregation of the type <<XX:XXxXX:XX>>. The allele separators (: in 
   the example) can be any of: :;,/| The meaning is similar to that read by
   ReadSegTypeCross above except that the alleles can be strings instead of
   single characters. At this point the characters << have already been read.
   
   The algorithm is identical to ReadSegTypeCross. The input format is more
   complicated and most of the function is error-checking on the input in an
   attempt to always output a helpful error message.
   
   This function inserts no more than 17 symbols into the symbol table.
 */
int ReadSegTypeLongCross(FILE *fp, outbredSymbol_t *symbols, int *pNumSymbs,
                         const char *markName)
{
    char    alleles[4][MAX_ALLELE_LEN];
    static int  *sepBase = NULL;
    int     *sep; /* lookup table for whether a character is a seperator */
    char    sepChar; /* the seperator character used by this cross */
    int     i;
    int     c;
    int     nullParentAllele = FALSE; /* is there a null allele in the cross? */
    
    /* This is allocated only once per run, instead of once per function call.
       This simplifies clean up at the many exit points. */
    if (sepBase == NULL)
        sepBase = (int *)malloc((CHAR_MAX - CHAR_MIN + 1) * sizeof(int));
    
    sep = sepBase - CHAR_MIN;
    for (i = CHAR_MIN; i <= CHAR_MAX; ++i)
        sep[i] = FALSE;
    sep[':'] = sep['/'] = sep[','] = sep[';'] = sep['|'] = TRUE; 
    
    /* read the alleles 4 different alleles that make up the cross */
    for (i = 0; i < 4; ++i) {
        int     j = 0;
        
        while (TRUE) {
            c = getc(fp);
            
            if (c == '>' && i == 3)
                break;
            
            if (c == EOF || c == '\n' || c == '>') {
                fprintf(stderr, "Error: marker %s: incomplete ", markName);
                fprintf(stderr, "segregation type\n");
                
                /* The calling function will read to the end of the line. If
                   we don't put the \n back in the file, it will skip the next
                   marker, which would be bad. */
                if (c == '\n')
                    ungetc(c, fp);
                return FALSE;
            }
            
            if (isspace(c)) {
                fprintf(stderr, "Error: marker %s: white space in ", markName);
                fprintf(stderr, "segregation type\n");
                return FALSE;
            }
            
            if (c == 'x') {
                if (i == 0) {
                    fprintf(stderr, "Error: marker %s has no ", markName);
                    fprintf(stderr, "allele separator before 'x'\n");
                    return FALSE;
                } else if (i == 1)
                    break;
                else {
                    fprintf(stderr, "Error: marker %s: too many x's\n",
                            markName);
                    return FALSE;
                }
            }
            
            if (sep[c]) {
                if (i == 0) {
                    sepChar = (char) c;
                    break;
                } else if (i == 2) {
                    if (c != sepChar) {
                        fprintf(stderr, "Error: marker %s: inconsistent ",
                                markName);
                        fprintf(stderr, "separators: %c and %c\n", sepChar, c);
                        return FALSE;
                    }
                    break;
                } else {
                    fprintf(stderr,"Error: marker %s: too many ", markName);
                    fprintf(stderr, "allele separators\n");
                    return FALSE;
                }
            }
            
            alleles[i][j++] = (char) c;
            
            if (j == MAX_ALLELE_LEN) {
                fprintf(stderr, "Error: marker %s: allele ", markName);
                fprintf(stderr, "beginning %.*s too long\n", j, alleles[i]);
                return FALSE;
            }
        }
        alleles[i][j] = '\0';
        
        if (strcmp(alleles[i], "0") == 0)
            nullParentAllele = TRUE;
    }
    
    /* This is the second of the two '>' characters which mark the end of the
       cross. If it is not present, just give a warning, since it doesn't cause
       any parsing ambiguities. */
    if ((c = getc(fp)) != '>') {
        fprintf(stderr, "Warning: marker %s: segregation type ", markName);
        fprintf(stderr, "should end with '>>'\n");
        ungetc(c, fp);
    }
    
    for (i = 0; i < 4; ++i) {
        char    *mothAllele = alleles[i % 2];
        char    *fathAllele = alleles[2 + (i / 2)];
        int     homozygous = strcmp(mothAllele, fathAllele) == 0;
        
        /* Sets up "-" + sepChar + "-" as an unknown genotype */
        AddLongCrossGenotype(symbols, pNumSymbs, "-", "-", sepChar, i);
        
        AddLongCrossGenotype(symbols, pNumSymbs, mothAllele, fathAllele,
                             sepChar, i);
        AddLongCrossGenotype(symbols, pNumSymbs, fathAllele, mothAllele,
                             sepChar, i);
        AddLongCrossGenotype(symbols, pNumSymbs, mothAllele, "?", sepChar, i);
        AddLongCrossGenotype(symbols, pNumSymbs, fathAllele, "?", sepChar, i);
        if (!nullParentAllele || homozygous || strcmp(fathAllele, "0") == 0)
            AddLongCrossGenotype(symbols, pNumSymbs, mothAllele, "-", sepChar,
                                 i);
        if (!nullParentAllele || homozygous || strcmp(mothAllele, "0") == 0)
            AddLongCrossGenotype(symbols, pNumSymbs, fathAllele, "-", sepChar,
                                 i);
    }
    
    return TRUE;
}

/* Takes a pair of alleles and a separator character, creates a name by
   concatenating these and puts it into the symbols table.
 */
void AddLongCrossGenotype(outbredSymbol_t *symbs, int *pCount, char *allele1,
                          char *allele2, char sep, int gen)
{
    int     len1 = strlen(allele1);
    int     len2 = strlen(allele2);
    char    *name = (char *)malloc(len1 + len2 + 2);
    
    memcpy(name, allele1, len1);
    name[len1] = sep;
    memcpy(name + len1 + 1, allele2, len2);
    name[len1 + len2 + 1] = '\0';
    
    AddSymbolGenotype(symbs, pCount, name, gen);
    free(name);
}

/* Reads the segregation type in the format:
       [<F0M0>,<F0M1>,<F1M0>,<F1M1>]
   The separation between the names and the brackets and the commas may be
   additionally separated by spaces. At the point in which it is called, the
   first bracket has already been parsed.
   
   This function adds a maximum of 4 symbols to the table.
 */
int ReadSegTypeExplicit(FILE *fp, outbredSymbol_t *symbols, int *pNumSymbs,
                         const char *markName)
{
    int         i, j;
    char        buf[BUF_SIZE];
    int         c;
    
    for (i = 0; i < 4; ++i) {
        while (isspace(c = getc(fp)))
            ;
        
        for (j = 0; !isspace(c) && c != ',' && c != ']'; ++j) {
            if (j >= BUF_SIZE - 1) {
                buf[BUF_SIZE - 1] = '\0';
                fprintf(stderr, "Error: phenotype name beginning with %s used",
                        buf);
                fprintf(stderr, " with marker %s is too long\n", markName);
                return FALSE;
            }
            
            buf[j] = c;
            c = getc(fp);
        }
        
        if (j == 0) {
            fprintf(stderr, "Error with segregation type of marker %s: ",
                    markName);
            fprintf(stderr, "blank phenotype name\n");
            return FALSE;
        }
        
        buf[j] = '\0';
        AddSymbolGenotype(symbols, pNumSymbs, buf, i);
        
        while (isspace(c))
            c = getc(fp);
        
        if (i < 3 && c != ',') {
            if (c == ']') {
                fprintf(stderr, "Error with segregation type of marker %s:\n",
                        markName);
                fprintf(stderr, "Should be 4 phenotypes, but only found %d\n",
                        i + 1);
            } else {
                fprintf(stderr, "Error with segregation type of marker %s:\n",
                        markName);
                fprintf(stderr, "Must use comma to separate phenotype names\n");
            }
            return FALSE;
        }
    }
    
    if (c != ']') {
        if (c == ',') {
            fprintf(stderr, "Error with segregation type of marker %s:\n",
                    markName);
            fprintf(stderr,"Can only have 4 phenotypes between the brackets\n");
        } else {
            fprintf(stderr, "Error with segregation type of marker %s:\n",
                    markName);
            fprintf(stderr, "Must have right bracket (]) at end of");
            fprintf(stderr, " segregation type description\n");
        }
        return FALSE;
    }
    
    return TRUE;
}

/* Reads the segregation type in the format:
     <F0M0> <F0M1> <F1M0> <F1M1> |
   Each of the FXMX is denotes a symbol corresponding to the phenotype which
   would be observed for that inheritance type.
   
   This function adds a maximum of 4 symbols to the table.
 */
int ReadSegTypeExplicitOld(FILE *fp, outbredSymbol_t *symbols, int *pNumSymbs,
                            const char *markName)
{
    char        tok[BUF_SIZE];
    int         i;
    
    for (i = 0; i < 4; ++i) {
        if (!ReadToken(tok, BUF_SIZE, fp)) {
            fprintf(stderr, "Error: marker %s: incomplete segregation type\n",
                    markName);
            return FALSE;
        }
        AddSymbolGenotype(symbols, pNumSymbs, tok, i);
    }
    
    if (!ReadToken(tok, BUF_SIZE, fp)) {
        fprintf(stderr, "Error: marker %s: incomplete segregation type\n",
                markName);
        return FALSE;
    }
    
    if (strcmp(tok, "|") != 0) {
        fprintf(stderr, "Error: marker %s: expected |, but found %s\n",
                markName, tok);
        return FALSE;
    }
    
    return TRUE;
}

/* Associates the specified symbol name with the genotype gen. If the name
   exists, the genotype is added to the current genotypes. If not, an entry is
   created.
 */
void AddSymbolGenotype(outbredSymbol_t *symbs, int *pCount, const char *name,
                       int gen)
{
    outbredSymbol_t *symb;
    
    symb = LookupOutbredSymbol(symbs, *pCount, name);
    if (symb == NULL) {
        symb = symbs + (*pCount)++;
        symb->name = (char *)malloc(strlen(name) + 1);
        strcpy(symb->name, name);
        symb->val = 0;
        symb->fixed = FALSE;
    }
    symb->val = (obs_t) ((1 << gen) | symb->val);
}

outbredSymbol_t *LookupOutbredSymbol(outbredSymbol_t *symbs, int count,
                                     const char *name)
{
    int     i;
    
    for (i = 0; i < count; ++i) {
        if (strcmp(name, symbs[i].name) == 0) {
            return symbs + i;
        }
    }
    return NULL;
}

void FreeOutbredSymbols(outbredSymbol_t *symbs, int numSymbs)
{
    int     i;
    
    for (i = 0; i < numSymbs; ++i) {
        if (!symbs[i].fixed)
            free(symbs[i].name);
    }
}

/* This function reads a single token from the input stream. However, it will
   not read past a newline, and if it reads a newline it puts it back in the
   buffer If there is a newline before the next non-whitespace character, then
   it will return FALSE and leave the file pointer just before the newline.
   Otherwise, it returns TRUE and writes the token to buf.
 */
int ReadToken(char *buf, int bufSz, FILE *fp)
{
    int     c;
    int     i;
    
    do
        c = getc(fp);
    while (isspace(c) && c != '\n');
    
    if (c == '\n') {
        ungetc(c, fp);
        return FALSE;
    }
    
    for (i = 0; i < bufSz - 1; ++i) {
        buf[i] = c;
        c = getc(fp);
        if (isspace(c) || c == EOF) {
            if (c == '\n')
                ungetc(c, fp);
            buf[i + 1] = '\0';
            return TRUE;
        }
    }
    
    buf[bufSz - 1] = '\0';
    fprintf(stderr, "Token beginning with %s in data file too long\n", buf);
    
    return TRUE;
}

void ReadToEol(FILE *fp)
{
    int     c;
    
    while ((c = getc(fp)) != '\n' && c != EOF)
        ;
}
