/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "mapPool.h"
#include <stdlib.h>

/* These functions deal with "map pools", which are arrays of maps with certain
   fixed capacities. When a new map is added which would exceed the capacity,
   the lowest likelihood map is removed. Thus, they keep track of a certain
   number of the highest likelihood maps at any given time.
 */

void allocMapPool(mapPool_t *pool, int capacity)
{
    pool->maps = (map_t *)malloc(capacity * sizeof(map_t));
    pool->size = 0;
    pool->capacity = capacity;
}

void clearMapPool(mapPool_t *pool)
{
    int     i;
    
    for (i = 0; i < pool->size; ++i)
        freeMap(pool->maps + i);
    pool->size = 0;
}

void freeMapPool(mapPool_t *pool)
{
    clearMapPool(pool);
    free(pool->maps);
}

void swapMapPools(mapPool_t *a, mapPool_t *b)
{
    map_t   *tmpMap;
    int     tmp;
    
    tmp = a->capacity;
    a->capacity = b->capacity;
    b->capacity = tmp;
    
    tmp = a->size;
    a->size = b->size;
    b->size = tmp;
    
    tmpMap = a->maps;
    a->maps = b->maps;
    b->maps = tmpMap;
}

/* Inserts a new map into the pool. Note that the size cannot be bigger than the
   smallest map already in the pool. Only use pools which have maps all of the
   same length.
 */
void insertMapPool(mapPool_t *pool, map_t *map)
{
    if (pool->size < pool->capacity){
        map_t   *dst = pool->maps + pool->size++;
        
        allocSimilarMap(dst, map);
        copyMap(dst, map);
    } else {
        int     i;
        int     ndx;
        prob_t  lowestLlhd = 0;
        
        for (i = 0; i < pool->size; ++i) {
            if (pool->maps[i].logLlhd < lowestLlhd) {
                ndx = i;
                lowestLlhd = pool->maps[i].logLlhd;
            }
        }
        
        if (map->logLlhd > lowestLlhd)
            copyMap(pool->maps + ndx, map);
    }
}

/* Returns a pointer to a map in the pool, which may not be valid after the
   pool is modified.
 */
map_t *bestMapPool(mapPool_t *pool)
{
    int     i;
    map_t   *res = pool->maps;
    
    for (i = 1; i < pool->size; ++i) {
        if (pool->maps[i].logLlhd > res->logLlhd)
            res = pool->maps + i;
    }
    
    return res;
}
