/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "genmap.h"
#include "ped.h"
#include "input.h"
#include "output.h"
#include <stdio.h>
#include <stdlib.h>

void project(pedigree_t *ped, int parent, const char *fname);

int main(int argc, char **argv)
{
    pedigree_t  ped;
    
    if (argc != 4) {
        fprintf(stderr, "Usage: %s <input> <out mother> <out father>\n", *argv);
        return EXIT_FAILURE;
    }
    
    if (!readFile(&ped, argv[1]))
        return EXIT_FAILURE;
    
    project(&ped, 0, argv[2]);
    project(&ped, 1, argv[3]);
    
    freePed(&ped);
}

/* Project onto one parent and duplicates it to have both possible phases. One
   of the copies has a '!' appended to its name.
     parent = 0 is mother
     parent = 1 is father. */
void project(pedigree_t *ped, int parent, const char *fname)
{
    FILE        *fp = fopen(fname, "w");
    const char  genChars[] = "0AH-";
    const char  genCharsSwap[] = "0HA-";
    char        *buf, *bufSwap;
    long        offStartCnt, offEndCnt;
    int         numInform = 0;
    
    if (fp == NULL) {
        fprintf(stderr, "Couldn't open %s for writing\n", fname);
        return;
    }
    
    buf = (char *) malloc(ped->numInds * sizeof(char));
    bufSwap = (char *) malloc(ped->numInds * sizeof(char));
    
    /* Cannot output the correct number of markers because we don't know
       yet how many will be informative in this parent. Thus, we output the
       total number of markers, and then come back at the end to fix it.
       We know that the number of informative markers will be less than the
       number of markers, and so the former will require no more characters to
       print. */
    fprintf(fp, "data type f2 backcross\n%d ", ped->numInds);
    offStartCnt = ftell(fp);
    fprintf(fp, "%d\n", ped->numMarks * 2);
    offEndCnt = ftell(fp) - 1; // - 1 because of the \n
    
    for (int i = 0; i < ped->numMarks; ++i) {
            /* Whether this marker is informative in this parent. */
        bool    informative = false;
        
        for (int j = 0; j < ped->numInds; ++j) {
            obs_t   gen = ped->markers[i].obs[j];
            int     partGen;
            
            if (parent) {
                int     a = (gen | (gen >> 1));
                partGen = ((a & 0x4) >> 1) | (a & 0x1);
            } else
                partGen = (gen | (gen >> 2)) & 0x3;
            
            buf[j] = genChars[partGen];
            bufSwap[j] = genCharsSwap[partGen];
            if (partGen != 0x3)
                informative = true;
        }
        
        if (informative) {
            fprintf(fp, "*%s ", ped->markers[i].name);
            fwrite(buf, 1, ped->numInds, fp);
            
            fprintf(fp, "\n*%s! ", ped->markers[i].name);
            fwrite(bufSwap, 1, ped->numInds, fp);
            fputc('\n', fp);
            
            ++numInform;
        }
    }
    free(buf);
    free(bufSwap);
    
    fseek(fp, offStartCnt, SEEK_SET);
    fprintf(fp, "%d", numInform * 2);
    /* We're overwriting a string of length offEndCnt - offStartCnt, and so
       we need to overwrite the whole string with spaces. */
    for (long i = offEndCnt - ftell(fp); i > 0; --i)
        putc(' ', fp);
    
    fclose(fp);
}
