/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "genmap.h"
#include "output.h"
#include "ped.h"
#include "map.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

void simPedigree(pedigree_t *ped, int numInd, int markPerLg, int numLg);
void unphase(pedigree_t *ped);

int main(int argc, char **argv)
{
    pedigree_t  ped;
    int         numInd, markPerLg, numLg;
    
    if (argc != 6) {
        fprintf(stderr, "usage: %s <num progeny> <mark per lg> ", *argv);
        fprintf(stderr, "<num lg> <phased out> <unphased out>\n");
        
        return EXIT_FAILURE;
    }
    
    numInd = atoi(argv[1]);
    markPerLg = atoi(argv[2]);
    numLg = atoi(argv[3]);
    
    simPedigree(&ped, numInd, markPerLg, numLg);
    writePed(&ped, argv[4]);
    unphase(&ped);
    writePed(&ped, argv[5]);
    
    freePed(&ped);
    return EXIT_SUCCESS;
}

/* Simulates a pedigree with mixed segregation types. It consists of numLg
   linkage groups, each with markPerLg markers. The markers are arranged in the
   output file first by linkage group and then by position within the linkage
   group.
 */
void simPedigree(pedigree_t *ped, int numInd, int markPerLg, int numLg)
{
    double  centiMorganPerLg = 100.0;
    double  missingProb = 0.02; /* proability of missing a given observation */
    double  errorProb = 0.01; /* probability of an erroneous genotype */
                /* Recombination fraction between markers in the same LG */
    double  thetaNormal = dist2Theta(centiMorganPerLg / markPerLg);
    int     numMark = markPerLg * numLg;
    obs_t   markTypes[][4] = {{0x1, 0x2, 0x4, 0x8}, /* fully informative */
                              {0x3, 0x3, 0xc, 0xc}, /* aaxab */
                              {0x5, 0xa, 0x5, 0xa}, /* abxaa */
                              {0x1, 0x6, 0x6, 0x8}, /* abxab */
                              {0x9, 0x2, 0x4, 0x9}, /* abxab */
                              {0x7, 0x7, 0x7, 0x8}, /* a0xa0 */
                              {0xb, 0xb, 0x4, 0xb}, /* a0xa0 */
                              {0xd, 0x2, 0xd, 0xd}, /* a0xa0 */
                              {0x1, 0xe, 0xe, 0xe}}; /* a0xa0 */
#if 1
    double  markTypeProb[] = {0.1, 0.25, 0.25, 0.125, 0.125,
                              0.0375, 0.0375, 0.0375, 0.0375};
#else
    double  markTypeProb[] = {0.15, 0.25, 0.25, 0.175, 0.175,
                              0.0, 0.0, 0.0, 0.0};
#endif
    int     numMarkType = sizeof(markTypes) / sizeof(*markTypes);
    obs_t   *obs;
    int     i, j;
    int     lg;
    int     *ibd;
    
    srand48(time(NULL));
    
    ped->numMarks = numMark;
    ped->numInds = numInd;
    ped->markers = (marker_t *)malloc(numMark * sizeof(marker_t));
    obs = (obs_t *)malloc(numMark * numInd * sizeof(obs_t));
    ibd = (int *)malloc(numInd * sizeof(int));
    
    /* Before storing the first marker, we will have a recombination with theta
       = 0.5, so the initial IBDs will effectively be random even though
       they're constant here. */
    for (i = 0; i < numInd; ++i)
        ibd[i] = 0;
    
    for (i = 0, lg = 0; i < numMark; ++i) {
        double  theta;
        int     markType;
        double  r = drand48();
        
        /* Choose recombination fraction between this marker and the last */
        if (i % markPerLg == 0) {
            theta = 0.5; /* starts a new linkage group */
            ++lg;
        } else
            theta = thetaNormal;
        
        /* Choose marker type */
        for (markType = 0; markType < numMarkType - 1; ++markType) {
            if ((r -= markTypeProb[markType]) < 0.0)
                break;
        }
            
        /* Create entry in pedigree */
        ped->markers[i].name = (char *)malloc(15);
        sprintf(ped->markers[i].name, "MARKER%d.%d", lg, i);
        ped->markers[i].obs = obs + i * numInd;
        
        /* Simulate recombination, then store genotype in pedigree structure */
        for (j = 0; j < numInd; ++j) {
            obs_t   obs;
            
            if (drand48() < theta)
                ibd[j] ^= 1;
            if (drand48() < theta)
                ibd[j] ^= 2;
            
            if (drand48() < missingProb)
                obs = OBSERVATION_UNKNOWN;
            else if (drand48() < errorProb)
                obs = markTypes[markType][lrand48() / (1 << 29)];
            else
                obs = markTypes[markType][ibd[j]];
            ped->markers[i].obs[j] = obs;
        }
    }
    
    free(ibd);
}

/* Randomizes the phases of all markers in the pedigree. */
void unphase(pedigree_t *ped)
{
    int     i;
    
    for (i = 0; i < ped->numMarks; ++i) {
        if (drand48() < 0.5)
            swapMaternalPhase(ped->markers + i, ped->numInds);
        if (drand48() < 0.5)
            swapPaternalPhase(ped->markers + i, ped->numInds);
    }
}
